import inspect
import logging

import opentracing

from opentracing import child_of, follows_from
from opentracing.ext import tags as opentracing_tags


DEFAULT_SPAN_ARG_NAME = '__OPENTRACINGUTILS_SPAN'  # hmmm!


logger = logging.getLogger(__name__)


def get_new_span(
        f, func_args, func_kwargs, operation_name=None, inspect_stack=True, ignore_parent_span=False,
        span_extractor=None, use_follows_from=False):
    parent_span = None
    span_arg_name = DEFAULT_SPAN_ARG_NAME

    if not ignore_parent_span:
        if callable(span_extractor):
            try:
                parent_span = span_extractor(*func_args, **func_kwargs)
            except Exception:
                logger.exception('Failed to extract span from: {}'.format(span_extractor.__name__))

        if not parent_span:
            span_arg_name, parent_span = get_parent_span(inspect_stack=inspect_stack, **func_kwargs)

    op_name = f.__name__ if not operation_name else operation_name

    references = None
    if parent_span:
        references = [follows_from(parent_span.context)] if use_follows_from else [child_of(parent_span.context)]

    return span_arg_name, opentracing.tracer.start_span(operation_name=op_name, references=references)


def adjust_span(span, operation_name, component, tags):
    if operation_name:
        span.set_operation_name(operation_name)

    if tags and type(tags) is dict:
        for k, v in tags.items():
            try:
                span.set_tag(k, v)
            except Exception:
                pass

    if component:
        span.set_tag(opentracing_tags.COMPONENT, component)

    return span


def get_span_from_kwargs(**kwargs):
    for k, v in kwargs.items():
        if isinstance(v, opentracing.Span):
            return k, v

    return None, None


def inspect_span_from_stack(depth=100):
    cframe = inspect.currentframe()
    frame = cframe.f_back

    span = None
    for _ in range(depth):
        if span or not frame:
            break

        for k, v in frame.f_locals.items():
            if isinstance(v, opentracing.Span):
                span = v

        frame = frame.f_back

    return span


def get_parent_span(inspect_stack=True, **kwargs):
    span_arg_name, parent_span = get_span_from_kwargs(**kwargs)

    if not parent_span and inspect_stack:
        span_arg_name = DEFAULT_SPAN_ARG_NAME
        parent_span = inspect_span_from_stack()

    return span_arg_name, parent_span


def extract_span_from_kwargs(**kwargs):
    """Return current span from kwargs"""
    _, span = get_span_from_kwargs(**kwargs)

    if not span:
        span = opentracing.tracer.start_span()
        span.log_kv({'opentracing_utils.warning': 'Cannot find span in kwargs. Starting new span!'})

    return span


def remove_span_from_kwargs(**kwargs):
    span_key, _ = get_span_from_kwargs(**kwargs)
    kwargs.pop(span_key, None)
    return kwargs



from __future__ import absolute_import
from future import standard_library
standard_library.install_aliases()  # noqa

import logging
import re
import urllib.parse as parse

try:
    import requests
except ImportError:  # pragma: no cover
    pass
else:
    __requests_http_send = requests.adapters.HTTPAdapter.send

import opentracing
from opentracing import Format
from opentracing.ext import tags as ot_tags

from opentracing_utils.decorators import trace
from opentracing_utils.span import get_span_from_kwargs
from opentracing_utils.common import sanitize_url


OPERATION_NAME_PREFIX = 'requests.send'

logger = logging.getLogger(__name__)


def trace_requests(default_tags=None, set_error_tag=True, mask_url_query=True,
                   mask_url_path=False, ignore_url_patterns=None):
    """Patch requests library with OpenTracing support.

    :param default_tags: Default span tags to included with every outgoing request.
    :type default_tags: dict

    :param set_error_tag: Set error tag to span if request is not ok.
    :type set_error_tag: bool

    :param mask_url_query: Mask URL query args.
    :type mask_url_query: bool

    :param mask_url_path: Mask URL path.
    :type mask_url_path: bool

    :param ignore_url_patterns: Ignore tracing for any URL's that match entries in this list
    :type ignore_url_patterns: list
    """
    def skip_span_matcher(http_adapter_obj, request, **kwargs):
        if ignore_url_patterns is None:
            return False

        if any(re.match(p, request.url) for p in ignore_url_patterns):
            return True

        return False

    @trace(pass_span=True, tags=default_tags, skip_span=skip_span_matcher)
    def requests_send_wrapper(self, request, **kwargs):
        if ignore_url_patterns is not None:
            if any(re.match(pattern, request.url) for pattern in ignore_url_patterns):
                return __requests_http_send(self, request, **kwargs)

        op_name = '{}.{}'.format(OPERATION_NAME_PREFIX, request.method)

        k, request_span = get_span_from_kwargs(inspect_stack=False, **kwargs)
        kwargs.pop(k, None)

        components = parse.urlsplit(request.url)

        if request_span:
            (request_span
                .set_operation_name(op_name)
                .set_tag(ot_tags.PEER_HOSTNAME, components.hostname)
                .set_tag(
                    ot_tags.HTTP_URL,
                    sanitize_url(request.url, mask_url_query=mask_url_query, mask_url_path=mask_url_path))
                .set_tag(ot_tags.HTTP_METHOD, request.method)
                .set_tag(ot_tags.SPAN_KIND, ot_tags.SPAN_KIND_RPC_CLIENT))

            # Inject our current span context to outbound request
            try:
                carrier = {}
                opentracing.tracer.inject(request_span.context, Format.HTTP_HEADERS, carrier)
                request.headers.update(carrier)
            except opentracing.UnsupportedFormatException:
                logger.error('Failed to inject span context in request!')

            resp = __requests_http_send(self, request, **kwargs)
            request_span.set_tag(ot_tags.HTTP_STATUS_CODE, resp.status_code)

            if set_error_tag and not resp.ok:
                request_span.set_tag('error', True)

            return resp
        else:
            logger.warn('Failed to extract span during initiating request!')
            return __requests_http_send(self, request, **kwargs)

    # The Patch!
    requests.adapters.HTTPAdapter.send = requests_send_wrapper
